#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 1500;
  nx = 301;
  nz = 301;
  ns = 10;
  sx = 100;
  sz = nz/2;
  boundary = 10;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat3D dspace(ns,nt,nx);         // space for data space
  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *u2     = new Vector(&uspace);    u2->zeroElement();       // Wave field  
  Vector *u3     = new Vector(&uspace);    u3->zeroElement();       // Wave field  

  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  SP->SetShot(4);
  SP->DotTest(true); 
  

  // Read in source wavelet and velocity field
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->Model1(K,rho);


  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  // Generate operators for wave progation
  Prop *Lmig = new Prop(&uspace,&uspace);
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  //Lmig->DotTest(true);
  
  // Inject source wavelet
  Source *Ss = new Source(nt, nz, nx);
  Ss->Injection(shot, u1, sz, sx);
  
  // Forward propagation
  double t0 = timer();
  Lmig->Forward(false,u1,u3);         // source -> ufield
  double t1 = timer(); 
  cout << "   wave propagation, takes " << t1-t0 << "seconds. " <<endl;
  
  float *__restrict__ u1_val = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();   
  float *__restrict__ u2_val = ((dataInCoreFloat*)u2->getDataContainer())->getFloatPtr();  
  float *__restrict__ u3_val = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();
  float *__restrict__ gk_val = ((dataInCoreFloat*)GradK->getDataContainer())->getFloatPtr();  
  float *__restrict__ gr_val = ((dataInCoreFloat*)Gradr->getDataContainer())->getFloatPtr();
  u2->zeroElement();
  for(llong i=0; i<nt;i++){
    u2_val[i*nx*nz+sz*nx+nx-sx] = u3_val[i*nx*nz+sz*nx+nx-sx];
   }  
  //Ss->BInjection(shot, u2, sz, nx-sx);
  Lmig->Adjoint(false,u1,u2);
  Lmig->kernel_K(u3,u1,GradK);
  Lmig->kernel_r(u3,u1,Gradr);     
  std::cout << "Grad r norm = " << Gradr->innerProd(*Gradr) << std::endl;
  srite("wave1",u3_val,sizeof(float)*nt*nz*nx);
  srite("wave2",u1_val,sizeof(float)*nt*nz*nx);
  srite("GradK",gk_val,sizeof(float)*nz*nx);
  srite("Gradr",gr_val,sizeof(float)*nz*nx);
  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}