#ifndef ENGINE_H
#define ENGINE_H 1
#include <cmath>
#include <vector>
#include "Taper.h"
#include "Clean.h"
#include <cmath>
#include <iostream>
#include "TimeMarch.h"
#include "base/utilities.h"
#include "Source.h"
#include "SensorPad.h"
#include "Prop.h"
#include "Imaging.h"
typedef long long llong;
/*
Born modeling operator
Need to set velocity before running
Assume source field ready to use.

Created:       2015-01-15
Last modified: 2015-01-19

*/

namespace SEP {
class Born: public Operator{
  public:
    Born(Space *_mspace, Space *_uspace){
      mspace = (inCoreFloat *)_mspace;
      uspace = (inCoreFloat *)_uspace;
      setDomain(_mspace);
      setRange(_uspace);
      std::vector<long long> out_dims = uspace->get_ns();
      nt = out_dims[0]; 
      nz = out_dims[1]; 
      nx = out_dims[2];
      temp = new double[nz*nx];
    }
    void SetPar(float _dt, float _dx){ dt = _dt; dx = _dx;}
    void SetBG(Vector *_BG) {BG = _BG;}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
  private:
    inCoreFloat *mspace, *uspace;
    llong nx, nz, nt;
    Vector *BG;
    float dt, dx;
    double *temp;
};

class BornXZ: public Operator{
  public:
    BornXZ(Space *_mspace, Space *_uspace){
      mspace = (inCoreFloat *)_mspace;
      uspace = (inCoreFloat *)_uspace;
      setDomain(_mspace);
      setRange(_uspace);
      std::vector<long long> out_dims = uspace->get_ns();
      nt = out_dims[0]; 
      nz = out_dims[1]; 
      nx = out_dims[2];
      temp = new double[2*nz*nx];
    }
    void SetPar(float _dt, float _dx, llong _nb){
     dt = _dt; dx = _dx; nb=_nb;
      tempx = new Vector(mspace);
      tempz = new Vector(mspace);  
      tempuf= new Vector(mspace);
      tempub= new Vector(mspace); 
      temp1 = new Vector(mspace);
      temp2 = new Vector(mspace);
      interpX = new IntX(mspace,mspace);
      interpZ = new IntZ(mspace,mspace);
 	  DX = new GradX(mspace,mspace, nb);          // Dx from regular to staggered
      DZ = new GradZ(mspace,mspace, nb);          // Dz from regular to staggered
      DX2 = new GradX2(mspace,mspace, nb);        // Dx from staggered to regular
      DZ2 = new GradZ2(mspace,mspace, nb);        // Dz from staggered to regular  
     }
    void SetBG(Vector *_BG) {BG = _BG;}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
  private:
    inCoreFloat *mspace, *uspace;
    llong nx, nz, nt;
    Vector *BG;
    float dt, dx;
    llong nb;
    double *temp;
    Vector *tempx, *tempz, *tempuf, *tempub, *temp1, *temp2;
	IntX *interpX;
	IntZ *interpZ;
    GradX *DX;
    GradZ *DZ;
    GradX2 *DX2;
    GradZ2 *DZ2;
};


class BornXZT: public Operator{
  public:
    BornXZT(Space *_mspace, Space *_uspace){
      mspace = (inCoreFloat *)_mspace;
      uspace = (inCoreFloat *)_uspace;
      setDomain(_mspace);
      setRange(_uspace);
      std::vector<long long> out_dims = uspace->get_ns();
      nt = out_dims[0]; 
      nz = out_dims[1]; 
      nx = out_dims[2];
      temp = new double[2*nz*nx];
    }
    void SetPar(float _dt, float _dx, llong _nb){
     dt = _dt; dx = _dx; nb=_nb;
      tempx = new Vector(mspace);
      tempz = new Vector(mspace);  
      tempuf= new Vector(mspace);
      tempub= new Vector(mspace); 
      temp1 = new Vector(mspace);
      temp2 = new Vector(mspace);
      interpX = new IntX(mspace,mspace);
      interpZ = new IntZ(mspace,mspace);
 	  DX = new GradX(mspace,mspace, nb);          // Dx from regular to staggered
      DZ = new GradZ(mspace,mspace, nb);          // Dz from regular to staggered
      DX2 = new GradX2(mspace,mspace, nb);        // Dx from staggered to regular
      DZ2 = new GradZ2(mspace,mspace, nb);        // Dz from staggered to regular  
     }
    void SetBG(Vector *_BG) {BG = _BG;}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
  private:
    inCoreFloat *mspace, *uspace;
    llong nx, nz, nt;
    Vector *BG;
    float dt, dx;
    llong nb;
    double *temp;
    Vector *tempx, *tempz, *tempuf, *tempub, *temp1, *temp2;
	IntX *interpX;
	IntZ *interpZ;
    GradX *DX;
    GradZ *DZ;
    GradX2 *DX2;
    GradZ2 *DZ2;
};



class Engine{
  
  public:
    Engine(Space *_mspace, Space *_dspace, Space *_uspace, Vector * _shot, std::vector<int> _sxv, std::vector<int> _szv){
      mspace = (inCoreFloat *)_mspace;
      dspace = (inCoreFloat *)_dspace;
      uspace = (inCoreFloat *)_uspace;
      shot = new Vector(_shot); sxv = _sxv; szv = _szv;
    }

    void SetPropPar(llong _nb, llong _bsize, llong _boundary, float _dt, float _dx, float _dz){
      nb = _nb; bsize = _bsize;
      boundary = _boundary;
      dt = _dt; dx = _dx; dz = _dz;
      init();
    }
    void SetModel(Vector *K, Vector *rho){
      for (int i=0; i < nb; i++){
        LVec[i]->SetModel(K, rho);
        //     std::cout << " Prop test: " << std::endl;
        //    LVec[i]->Dot();
      }     
    }
    void CreateData();
    void GradientFunc(Vector *Kmodel, Vector *rmodel, Vector *K0, Vector *r0);
    void HessianFunc(Vector *K0, Vector *r0, Vector *Kmodel, Vector *rmodel, Vector *HK, Vector *Hrho);
    void GaussHessianFunc(Vector *K0, Vector *r0, Vector *Kmodel, Vector *rmodel, Vector *HK, Vector *Hrho);
    void GetData(float *data);
    void PutData(float *data);
    void GetGrad(float *gradk);
    void GetWave(float *wave, llong is){
      float *__restrict__ u_val = ((dataInCoreFloat*)UVec[is]->getDataContainer())->getFloatPtr(); 
      for(llong i=0; i < nx*nz*nt; i++){
        wave[i] = u_val[i];
      }
    
    }
    ~Engine(){}
    
  private:
    void init();
    llong ns, nb, bsize;
    llong sx,sz;
    llong nz,nx,nt, boundary;
    std::vector<int> sxv, szv;
    float dt, dx, dz;
    float alpha,pi;
    Vector *K, *rho, *shot, *acq;
    inCoreFloat *mspace, *dspace;
    inCoreFloat *uspace;
    std::vector<Vector *> DataVec, Data1Vec;
    std::vector<Vector *> dKVec,drVec, dmK,dmr;
    std::vector<Vector *> UVec, SVec, VVec,ShotVec, U1Vec,V1Vec;
    std::vector<Prop *> LVec;
    std::vector<Source *>  SSVec;
    std::vector<SensorPad *> PadVec;
    //std::vector<Gradient *> GradVec;
    std::vector<Born *> BornVec;
    std::vector<BornXZ *> BornXZVec;
    std::vector<BornXZT *> BornXZTVec;
    //std::vector<BornF *> BornFVec;
    void CombineM(std::vector<Vector *>& MV, Vector *model, Vector *model0);
    void CombineMH(std::vector<Vector *>& mV, Vector *model);
    CleanVec Zero;
};




bool Born::Forward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  llong mblock = nx*nz; 
  double dt2 = 1./(dt*dt*nt);
  for(long long i = 0; i < nz*nx; i++) {
	  d[i] = 0;    //      u field
  }
  for(llong i = 0; i < nz*nx; i++) temp[i] = m[i];
  for(llong it=1; it < nt-1; it++){
	llong shift = it*mblock;
	for(long long i = 0; i < nz*nx; i++) {
	  d[i + shift] += dt2*temp[i]*(-1.*uval[i + shift-mblock]+2.*uval[i + shift]-uval[i + shift+mblock]);    //      u field
	}
  } 
} 

bool Born::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  llong mblock = nx*nz; 
  double dt2 = 1./(dt*dt*nt);
  for(llong i = 0; i < nz*nx; i++) temp[i] = 0.;
  for(llong it=1; it < nt-1; it++){
	llong shift = it*mblock;
	for(long long i = 0; i < nz*nx; i++) {
	  temp[i] += dt2*d[i + shift]*(-1.*uval[i + shift-mblock]+2.*uval[i + shift]-uval[i + shift+mblock]);    //      u field
	}
  }     
  for(llong i = 0; i < nz*nx; i++) m[i] += temp[i];
}

bool BornXZ::Forward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  llong mblock = nx*nz; 
  double dx2 = 1./(dx*dx*nt);
  interpX->Forward(false,model,tempx);
  interpZ->Forward(false,model,tempz); 
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX->Forward(false,tempuf,temp1);
    DZ->Forward(false,tempuf,temp2);
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      u field
      temp2val[i] = temp2val[i]*tempzval[i];    //      u field
    }  
    DX2->Adjoint(false,tempuf,temp1);
    DZ2->Adjoint(false,tempub,temp2);
    
    for(long long i = 0; i < nz*nx; i++) {
      dval[i + shift] += (tempufval[i]+tempubval[i])*dx2;    //      u field
    }
  }   
  
  
} 

bool BornXZ::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  llong mblock = nx*nz; 
  double dx2 = 1./(dx*dx*nt);
    for(long long i = 0; i < 2*nz*nx; i++) {
      temp[i]=0.;
    } 
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = dval[i + shift]*dx2;    //      u field
    }
    
    DX2->Forward(false,tempuf,tempx);
    DZ2->Forward(false,tempuf,tempz);
    
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX->Forward(false,tempuf,temp1);
    DZ->Forward(false,tempuf,temp2);
    for(long long i = 0; i < nz*nx; i++) {
      tempxval[i] = temp1val[i]*tempxval[i];    //      u field
      tempzval[i] = temp2val[i]*tempzval[i];    //      u field
    }  
    
    for(long long i = 0; i < nz*nx; i++) {
      temp[i] += tempxval[i];    //      u field
      temp[i+mblock] += tempzval[i];
    }    
 
  } 
    for(long long i = 0; i < nz*nx; i++) {
      tempxval[i]= temp[i] ;    //      u field
      tempzval[i] = temp[i+mblock];
    }     
    interpX->Adjoint(true,model,tempx);
    interpZ->Adjoint(true,model,tempz);  
}


bool BornXZT::Forward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  llong mblock = nx*nz; 
  double dx2 = 1./(dx*dx*nt);
  interpX->Forward(false,model,tempx);
  interpZ->Forward(false,model,tempz); 
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX2->Forward(false,tempuf,temp1);
    DZ2->Forward(false,tempuf,temp2);
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      u field
      temp2val[i] = temp2val[i]*tempzval[i];    //      u field
    }  
    DX->Adjoint(false,tempuf,temp1);
    DZ->Adjoint(false,tempub,temp2);
    
    for(long long i = 0; i < nz*nx; i++) {
      dval[i + shift] += (tempufval[i]+tempubval[i])*dx2;    //      u field
    }
  }   
  
  
} 

bool BornXZT::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  llong mblock = nx*nz; 
  double dx2 = 1./(dx*dx*nt);
    for(long long i = 0; i < 2*nz*nx; i++) {
      temp[i]=0.;
    } 
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = dval[i + shift]*dx2;    //      u field
    }
    
    DX->Forward(false,tempuf,tempx);
    DZ->Forward(false,tempuf,tempz);
    
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX2->Forward(false,tempuf,temp1);
    DZ2->Forward(false,tempuf,temp2);
    for(long long i = 0; i < nz*nx; i++) {
      tempxval[i] = temp1val[i]*tempxval[i];    //      u field
      tempzval[i] = temp2val[i]*tempzval[i];    //      u field
    }  
    
    for(long long i = 0; i < nz*nx; i++) {
      temp[i] += tempxval[i];    //      u field
      temp[i+mblock] += tempzval[i];
    }    
 
  } 
    for(long long i = 0; i < nz*nx; i++) {
      tempxval[i]= temp[i] ;    //      u field
      tempzval[i] = temp[i+mblock];
    }     
    interpX->Adjoint(true,model,tempx);
    interpZ->Adjoint(true,model,tempz);  
}


void Engine::init() {
  std::cout<< "Initializing Born operator." <<std::endl;
  std::vector<long long> in_dims = mspace->get_ns();
  std::vector<long long> out_dims = dspace->get_ns();
  
  // Dimension of parameters m[nz][nx]  p[nt][nz][nx]
  nz = in_dims[0]; 
  nx = in_dims[1]; 
  ns = sxv.size();
  nt = out_dims[0];
  // Default parameters
  pi = 3.1415926535;

  Vector *temp = new Vector(mspace);
  // Create Propagation vector
  llong i;
  for (i=0; i < nb; i++){
    LVec.push_back(new Prop(uspace,uspace));
    drVec.push_back(new Vector(mspace));
    dKVec.push_back(new Vector(mspace));
    dmK.push_back(new Vector(mspace));
    dmr.push_back(new Vector(mspace));
    UVec.push_back(new Vector(uspace));
    U1Vec.push_back(new Vector(uspace));
    SVec.push_back(new Vector(uspace));
    VVec.push_back(new Vector(uspace));
    V1Vec.push_back(new Vector(uspace));
    ShotVec.push_back(new Vector(shot));
    LVec[i]->SetPropPar(boundary, dt, dx, temp);
    SSVec.push_back(new Source(nt, nz, nx));
    PadVec.push_back(new SensorPad(dspace, uspace, 60));
    //GradVec.push_back(new Gradient(UVec[i], temp, boundary, dt, dx));
    BornVec.push_back(new Born(mspace, uspace));
    BornVec[i]->SetPar(dt,dx);
    BornXZVec.push_back(new BornXZ(mspace, uspace));
    BornXZVec[i]->SetPar(dt,dx,boundary);
    BornXZTVec.push_back(new BornXZT(mspace, uspace));
    BornXZTVec[i]->SetPar(dt,dx,boundary);
    //BornFVec.push_back(new BornF(mspace, uspace,dspace));
    //BornFVec[i]->SetPar(dt,dx);
  }

  for(i=0; i < ns; i++){

    DataVec.push_back(new Vector(dspace));
    Data1Vec.push_back(new Vector(dspace));
  }
}



void Engine::GradientFunc(Vector *Kmodel, Vector *rmodel, Vector *K0, Vector *r0){

  int ib,id,is;
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    dKVec[ib]->zeroElement();
    drVec[ib]->zeroElement(); 
  }
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      // Source injection
      SVec[ib]->zeroElement();
      SSVec[ib]->Injection(ShotVec[ib], SVec[ib], szv[is], sxv[is]);
      // Forward simulation
      LVec[ib]->Forward(false,SVec[ib],UVec[ib]);       
      /*   
      std::cout << " Born test: " << std::endl;
        BornXZVec[ib]->SetBG(UVec[ib]);
        BornXZVec[ib]->DotTest(true);    
        */ 
      // Injection data residual     
      PadVec[ib]->SetShot(sxv[is]);
      PadVec[ib]->Forward(false,DataVec[is],SVec[ib]);   
      
      // Back simulation
      LVec[ib]->Adjoint(false,V1Vec[ib],SVec[ib]);
      
 
      // Imaging condition
      BornVec[ib]->SetBG(UVec[ib]);  
      BornVec[ib]->Adjoint(true,dKVec[ib],V1Vec[ib]);
      BornXZVec[ib]->SetBG(UVec[ib]);  
      BornXZVec[ib]->Adjoint(true,drVec[ib],V1Vec[ib]); 
  //    std::cout << "IS = " << is << ", dK norm = " << dKVec[is]->innerProd(*dKVec[is]) << std::endl;
    }
  }
  CombineM(dKVec, Kmodel,K0);  
  CombineM(drVec, rmodel,r0);
  //rmodel->zeroElement();   
}


void Engine::HessianFunc(Vector *K0, Vector *r0, Vector *dKmodel, Vector *drmodel, Vector *HK, Vector *Hrho){
  int ib,id,is;
  //drmodel->zeroElement();   // only use dK
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ K_val = ((dataInCoreFloat*)dKmodel->getDataContainer())->getFloatPtr();
    float *__restrict__ dmK_val = ((dataInCoreFloat*)dmK[ib]->getDataContainer())->getFloatPtr(); 
    float *__restrict__ rho_val = ((dataInCoreFloat*)drmodel->getDataContainer())->getFloatPtr();   
    float *__restrict__ dmrho_val = ((dataInCoreFloat*)dmr[ib]->getDataContainer())->getFloatPtr();
    float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr();
    float *__restrict__ r0_val = ((dataInCoreFloat*)r0->getDataContainer())->getFloatPtr();   
    for(llong i=0; i < nx*nz; i++){
      dmrho_val[i] = -rho_val[i]/(r0_val[i]*r0_val[i]);
      dmK_val[i] = -K_val[i]/(K0_val[i]*K0_val[i]);   
    }   
    dKVec[ib]->zeroElement();
    drVec[ib]->zeroElement();
  }

  //  std::cout <<" Entering !" <<std::endl;  
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      PadVec[ib]->SetShot(sxv[is]);
      
      // Source injection
      SVec[ib]->zeroElement();
      SSVec[ib]->Injection(ShotVec[ib], SVec[ib], szv[is], sxv[is]);
      
      // Forward simulation
      LVec[ib]->Forward(false,SVec[ib],UVec[ib]);        // U for background  
      
      // Back simulation
      PadVec[ib]->Forward(false,DataVec[is],SVec[ib]);   
      LVec[ib]->Adjoint(false,VVec[ib],SVec[ib]);        // VVec = lambda
      
      // Born approximation of Forward  
      BornVec[ib]->SetBG(UVec[ib]);  
      BornVec[ib]->Forward(false,dmK[ib],SVec[ib]);
      BornXZVec[ib]->SetBG(UVec[ib]);  
      BornXZVec[ib]->Forward(true,dmr[ib],SVec[ib]);         
      LVec[ib]->Forward(false,SVec[ib],U1Vec[ib]);
      
      // Born approxi + 2nd order scatter of Adjoint
      BornVec[ib]->SetBG(VVec[ib]);
      BornVec[ib]->Forward(false,dmK[ib],SVec[ib]);
      BornXZTVec[ib]->SetBG(VVec[ib]);  
      BornXZTVec[ib]->Forward(true,dmr[ib],SVec[ib]);   
      PadVec[ib]->Adjoint(false,Data1Vec[is],U1Vec[ib]);
      PadVec[ib]->Forward(true,Data1Vec[is],SVec[ib]);
      
      // Backward propagation
       LVec[ib]->Adjoint(false,V1Vec[ib],SVec[ib]);
       
      // Imaging condition
      BornVec[ib]->SetBG(UVec[ib]);  
      BornVec[ib]->Adjoint(true,dKVec[ib],V1Vec[ib]);
      BornXZVec[ib]->SetBG(UVec[ib]);  
      BornXZVec[ib]->Adjoint(true,drVec[ib],V1Vec[ib]);
      BornVec[ib]->SetBG(U1Vec[ib]);  
      BornVec[ib]->Adjoint(true,dKVec[ib],VVec[ib]);
      BornXZVec[ib]->SetBG(U1Vec[ib]);  
      BornXZVec[ib]->Adjoint(true,drVec[ib],VVec[ib]);  
    }
  }
  CombineM(dKVec, HK,K0); 
  CombineM(drVec, Hrho,r0); 
  //Hrho->zeroElement();
}


void Engine::GaussHessianFunc(Vector *K0, Vector *r0, Vector *dKmodel, Vector *drmodel, Vector *HK, Vector *Hrho){
  int ib,id,is;
  //drmodel->zeroElement();   // only use dK
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ K_val = ((dataInCoreFloat*)dKmodel->getDataContainer())->getFloatPtr();
    float *__restrict__ dmK_val = ((dataInCoreFloat*)dmK[ib]->getDataContainer())->getFloatPtr(); 
    float *__restrict__ rho_val = ((dataInCoreFloat*)drmodel->getDataContainer())->getFloatPtr();   
    float *__restrict__ dmrho_val = ((dataInCoreFloat*)dmr[ib]->getDataContainer())->getFloatPtr();
    float *__restrict__ K0_val = ((dataInCoreFloat*)K0->getDataContainer())->getFloatPtr();
    float *__restrict__ r0_val = ((dataInCoreFloat*)r0->getDataContainer())->getFloatPtr();   
    for(llong i=0; i < nx*nz; i++){
      dmrho_val[i] = -rho_val[i]/(r0_val[i]*r0_val[i]);
      dmK_val[i] = -K_val[i]/(K0_val[i]*K0_val[i]);   
    }   
    dKVec[ib]->zeroElement();
    drVec[ib]->zeroElement();
  }
  //  std::cout <<" Entering !" <<std::endl;  
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      PadVec[ib]->SetShot(sxv[is]);
      
      // Source injection
      SVec[ib]->zeroElement();
      SSVec[ib]->Injection(ShotVec[ib], SVec[ib], szv[is], sxv[is]);
      
      // Forward simulation
      LVec[ib]->Forward(false,SVec[ib],UVec[ib]);        // U for background  
       
      // Born approximation of Forward           
      BornVec[ib]->SetBG(UVec[ib]);  
      BornVec[ib]->Forward(false,dmK[ib],SVec[ib]);
      BornXZVec[ib]->SetBG(UVec[ib]);  
      BornXZVec[ib]->Forward(true,dmr[ib],SVec[ib]);      
      LVec[ib]->Forward(false,SVec[ib],U1Vec[ib]);
 
         
      // Born approxi + 2nd order scatter of Adjoint
      PadVec[ib]->Adjoint(false,Data1Vec[is],U1Vec[ib]);
      PadVec[ib]->Forward(false,Data1Vec[is],SVec[ib]);
  
      // Backward propagation
       LVec[ib]->Adjoint(false,V1Vec[ib],SVec[ib]);
    
      // Imaging condition
      BornVec[ib]->SetBG(UVec[ib]);  
      BornVec[ib]->Adjoint(true,dKVec[ib],V1Vec[ib]);
      BornXZVec[ib]->SetBG(UVec[ib]);  
      BornXZVec[ib]->Adjoint(true,drVec[ib],V1Vec[ib]); 
    }
  }
  CombineM(dKVec, HK,K0); 
  CombineM(drVec, Hrho,r0); 
  //Hrho->zeroElement();
}


void Engine::CreateData(){
  int ib,id,is;
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      //std::cout << "  Working at shot " << is << std::endl;
      SVec[ib]->zeroElement();
      SSVec[ib]->Injection(shot, SVec[ib], szv[is], sxv[is]);
      LVec[ib]->Forward(false,SVec[ib],UVec[ib]);
      PadVec[ib]->SetShot(sxv[is]);
      PadVec[ib]->Adjoint(false,DataVec[is],UVec[ib]);
//      std::cout << "is = " << is << ", sx = " << sxv[is] << ", sz = " << szv[is];
//      std::cout << ", norm = " << DataVec[is]->innerProd(*DataVec[is]) << std::endl;
    }
  }
}

void Engine::GetGrad(float *gradk){
  for(llong i=0; i < nb; i++){
  float *__restrict__ d_val = ((dataInCoreFloat*)dKVec[i]->getDataContainer())->getFloatPtr();
 // std::cout << "IS = " << i << ", dK norm = " << dKVec[i]->innerProd(*dKVec[i]) << std::endl;
    for(llong j=0; j < nz*nx; j++){
      gradk[i*nz*nx+j]=d_val[j];
    }
  }

}

void Engine::GetData(float *data){
  for(llong i=0; i < ns; i++){
  float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      data[i*nt*nx+j]=d_val[j];
    }
  }

}
void Engine::PutData(float *data){
  for(llong i=0; i < ns; i++){
  DataVec[i]->zeroElement();
  float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      d_val[j]=data[i*nt*nx+j];
    }
  }

}


void Engine::CombineM(std::vector<Vector *>& mV, Vector *model, Vector *model0){
  llong i,j;
  float *__restrict__ mval;
   float *__restrict__ mvec; 
  float *__restrict__ mval0;

  for(i = 0; i < nb; i++){
    mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
    mvec = ((dataInCoreFloat*)mV[i]->getDataContainer())->getFloatPtr();
    mval0 = ((dataInCoreFloat*)model0->getDataContainer())->getFloatPtr();
    for(j = 0; j < nz*nx; j++){
      mval[j] += -mvec[j]/(mval0[j]*mval0[j]);
    }
  }

}

void Engine::CombineMH(std::vector<Vector *>& mV, Vector *model){
  llong i,j;
  float *mval;
   float *mvec; 

  for(i = 0; i < nb; i++){
    mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
    mvec = ((dataInCoreFloat*)mV[i]->getDataContainer())->getFloatPtr();
    for(j = 0; j < nz*nx; j++){
      mval[j] += mvec[j];
    }
  }

}

}




#endif