#ifndef IMAGING_H
#define IMAGING_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include <omp.h>

#include "Interpolation.h"
#include "Grad.h"
typedef long long llong;
/*
RTM stacking Operator.
  Not really an operator. 
  There is dot test, but should be careful with omp

Created:       2015-01-15
Last modified: 2015-01-31

*/ 

namespace SEP {

class Gradient{
  public:
    Gradient(Vector *pressure, Vector *model, llong nb, float _dt, float _dx) {
      dt = _dt;
      dx = _dx;
      inCoreFloat *in_space = (inCoreFloat *)(pressure->getSpace());
      std::vector<long long> in_dims = in_space->get_ns();
      nt = in_dims[0];
      nz = in_dims[1];
      nx = in_dims[2];
      tempx = new Vector(model);
      tempz = new Vector(model);  
      tempuf= new Vector(model);
      tempub= new Vector(model); 
      temp  = new Vector(model);
      temp1 = new Vector(model);
      temp2 = new Vector(model);
      inCoreFloat *grad_space = (inCoreFloat *)(model->getSpace()); 
      interpX = new IntX(grad_space,grad_space);
      interpZ = new IntZ(grad_space,grad_space);
 	  DX = new GradX(grad_space,grad_space, nb);          // Dx from regular to staggered
      DZ = new GradZ(grad_space,grad_space, nb);          // Dz from regular to staggered
      DX2 = new GradX2(grad_space,grad_space, nb);        // Dx from staggered to regular
      DZ2 = new GradZ2(grad_space,grad_space, nb);        // Dz from staggered to regular         
    }
    Gradient(llong _nt, llong _nz, llong _nx){nt=_nt; nx=_nx; nz=_nz;}
    
    void GradK(Vector *uf, Vector *ub, Vector *GK);
    void Gradr(Vector *uf, Vector *ub, Vector *Gr);
    void BornS(Vector *dK,Vector *drho, Vector *u, Vector *s);
    void BornSAdj(Vector *dK,Vector *drho, Vector *u, Vector *s);
  private:
    llong nt,nx,nz;
    float dt, dx;
    Vector *tempx, *tempz, *tempuf, *tempub, *temp, *temp1, *temp2;
	IntX *interpX;
	IntZ *interpZ;
    GradX *DX;
    GradZ *DZ;
    GradX2 *DX2;
    GradZ2 *DZ2;
	void Gradrx(Vector *uf, Vector *ub, Vector *Grx);
	void Gradrz(Vector *uf, Vector *ub, Vector *Grz);
};


void Gradient::GradK(Vector *uf, Vector *ub, Vector *GK){
  float *__restrict__ uval = ((dataInCoreFloat*)uf->getDataContainer())->getFloatPtr();
  float *__restrict__ vval = ((dataInCoreFloat*)ub->getDataContainer())->getFloatPtr();
  float *__restrict__ Kval = ((dataInCoreFloat*)GK->getDataContainer())->getFloatPtr();
  float dt2=1./(dt*dt*nt);
  llong shift1 = 0;
  llong shift2 = 0;
  llong shift3 = 0;
  llong mblock = nx*nz;
  for(llong it = 1; it < nt - 1; it++){
    shift1 = mblock*(it-1);
    shift2 = mblock*it;
    shift3 = mblock*(it+1);
   // #pragma omp parallel for
    for(llong i = 0 ; i < mblock; i++){
      Kval[i] += -vval[shift2+i]*(uval[shift1+i]-2*uval[shift2+i]+uval[shift3+i])*dt2;
    }
  }
}

void Gradient::Gradr(Vector *uf, Vector *ub, Vector *Gr){
    tempx->zeroElement();
	Gradrx(uf,ub,tempx);
	interpX->Adjoint(true,Gr,tempx);
    tempz->zeroElement();
	Gradrz(uf,ub,tempz);
	interpZ->Adjoint(true,Gr,tempz);
}


void Gradient::Gradrx(Vector *uf, Vector *ub, Vector *Grx){
  float *__restrict__ uval = ((dataInCoreFloat*)uf->getDataContainer())->getFloatPtr();
  float *__restrict__ vval = ((dataInCoreFloat*)ub->getDataContainer())->getFloatPtr();
  float *__restrict__ rxval = ((dataInCoreFloat*)Grx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();
  float dx2=1./(dx*dx*nt);
  llong shift = 0;
  llong mblock = nx*nz;
  for(llong it=1; it<nt-1; it++){
    shift = it*mblock;
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift]*dx2;    //      u field
      tempubval[i] = vval[i + shift];    // lambda field
    }    
    DX->Forward(false,tempuf,temp1);
    DX2->Adjoint(false,temp2,tempub);
    for(llong i=0; i <mblock; i++){
      rxval[i] += temp1val[i]*temp2val[i];
    }
  }  // end of for it  

}  // end of Gradrx

void Gradient::Gradrz(Vector *uf, Vector *ub, Vector *Grz){
  float *__restrict__ uval = ((dataInCoreFloat*)uf->getDataContainer())->getFloatPtr();
  float *__restrict__ vval = ((dataInCoreFloat*)ub->getDataContainer())->getFloatPtr();
  float *__restrict__ rzval = ((dataInCoreFloat*)Grz->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();
  float dx2=1./(dx*dx*nt);
  llong shift = 0;
  llong mblock = nx*nz;
  for(llong it=1; it<nt-1; it++){
    shift = it*mblock;
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift]*dx2;    //      u field
      tempubval[i] = vval[i + shift];    // lambda field
    }    
    DZ->Forward(false,tempuf,temp1);
    DZ2->Adjoint(false,temp2,tempub);
    for(llong i=0; i <mblock; i++){
      rzval[i] += temp1val[i]*temp2val[i];
    }
  }  // end of for it  

}  // end of Gradrx

void Gradient::BornS(Vector *dmK,Vector *dmrho, Vector *u, Vector *s){
  float *__restrict__ dkval    = ((dataInCoreFloat*)dmK->getDataContainer())->getFloatPtr();
  float *__restrict__ drval    = ((dataInCoreFloat*)dmrho->getDataContainer())->getFloatPtr();
  float *__restrict__ uval     = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();
  float *__restrict__ sval     = ((dataInCoreFloat*)s->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ tempval  = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  s->zeroElement();
  interpX->Forward(false,dmrho,tempx);   // perturbation on the staggered grid
  interpZ->Forward(false,dmrho,tempz);   // perturbation on the staggered grid
   
  llong mblock = nx*nz; 
  double dt2 = 1./(dt*dt);
  double dx2 = 1./(dx*dx);
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX->Forward(false,tempuf,temp1);
    DZ->Forward(false,tempuf,temp2);
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      u field
      temp2val[i] = temp2val[i]*tempzval[i];    //      u field
    }   
    DX2->Adjoint(false,temp,temp1);
    DZ2->Adjoint(true,temp,temp2);
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      sval[i + shift] = tempval[i]*dx2+dt2*dkval[i]*(-1.*uval[i + shift-mblock]+2.*uval[i + shift]-uval[i + shift+mblock]);    //      u field
    } 
  //  std::cout << " It = " << it << " , born norm = " << s->innerProd(*s) << std::endl;   
  }  // end for for it

}  // End of Born operator

void Gradient::BornSAdj(Vector *dmK,Vector *dmrho, Vector *u, Vector *s){
  float *__restrict__ dkval    = ((dataInCoreFloat*)dmK->getDataContainer())->getFloatPtr();
  float *__restrict__ drval    = ((dataInCoreFloat*)dmrho->getDataContainer())->getFloatPtr();
  float *__restrict__ uval     = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();
  float *__restrict__ sval     = ((dataInCoreFloat*)s->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ tempval  = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempx->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempz->getDataContainer())->getFloatPtr();   
  
  interpX->Forward(false,dmrho,tempx);   // perturbation on the staggered grid
  interpZ->Forward(false,dmrho,tempz);   // perturbation on the staggered grid
   
  llong mblock = nx*nz; 
  double dt2 = 1./(dt*dt);
  double dx2 = 1./(dx*dx);
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }    
    DX2->Adjoint(false,temp1,tempuf);
    DZ2->Adjoint(false,temp1,tempuf);
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      u field
      temp2val[i] = temp2val[i]*tempzval[i];    //      u field
    }   
    DX->Forward(false,temp1,temp);
    DZ->Forward(true,temp2,temp);
    #pragma omp parallel for
    for(long long i = 0; i < nz*nx; i++) {
      sval[i + shift] = -tempval[i]*dx2+dt2*dkval[i]*(-1.*uval[i + shift-mblock]+2.*uval[i + shift]-uval[i + shift+mblock]);    //      u field
    } 
  //  std::cout << " It = " << it << " , born norm = " << s->innerProd(*s) << std::endl;   
  }  // end for for it

}  // End of Born operator
}
#endif