#ifndef LAPLACIAN_H
#define LAPLACIAN_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include <omp.h>
typedef long long llong;

/*
	Created by Yinbin Ma  
	4th order Laplacian in 2 d.
	Use carefully and have fun.
	
	First created: 2015-01-01
	Last Modified: 2015-02-28
*/	

namespace SEP {

class Laplacian4: public Operator{
  public:
    Laplacian4(Space *mod, Space *data){
      setDomain(mod);
      setRange(data);
      init();
    }
    ~Laplacian4(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    private:
    void init();
    float a0, a1, a2;
    float a0x,a1x,a2x,a3x,a4x,a0z,a1z,a2z,a3z,a4z;
    llong nz,nx;
    llong n1, n2;
    llong i1, i2, position;
    float temp0,temp1,temp2;
    CleanVec Zero;

};

void Laplacian4::init() {
  a0 = -5.;
  a1 = 4./3.;
  a2 = -1./12.;
  a0x = -205./72.;
  a1x = 8./5.;
  a2x = -1./5.;
  a3x = 8./315.;
  a4x = -1./560.;
  a0z = a0x;
  a1z = a1x;
  a2z = a2x;
  a3z = a3x;
  a4z = a4x;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n1 = in_dims[0];nx=n1;
  n2 = in_dims[1];nz=n2;
}

bool Laplacian4::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,n2,n1);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  llong ix, iz, position;
  for(ix = 4; ix < nx-4; ix++){
    for( iz = 4; iz < nz - 4; iz++) {
      position = ix * nz + iz;
      d[position] += 2.*a0x*m[position];
      d[position] += a1x*(m[position-nz*1]+m[position+nz*1]+m[position+1]+m[position-1]);
      d[position] += a2x*(m[position-nz*2]+m[position+nz*2]+m[position+2]+m[position-2]);
      d[position] += a3x*(m[position-nz*3]+m[position+nz*3]+m[position+3]+m[position-3]);
      d[position] += a4x*(m[position-nz*4]+m[position+nz*4]+m[position+4]+m[position-4]);      
    }
  }
  return 0;
}

bool Laplacian4::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n2,n1);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  llong ix, iz, position;
  for(ix = 4; ix < nx - 4; ix++){
    for(iz = 4; iz < nz - 4; iz++) {
      position = ix * nz + iz;
      m[position]      += a0x*d[position];
      m[position+nz*1] += a1x*d[position];
      m[position+nz*2] += a2x*d[position];
      m[position+nz*3] += a3x*d[position];  
      m[position+nz*4] += a4x*d[position];  
      m[position-nz*1] += a1x*d[position];
      m[position-nz*2] += a2x*d[position];
      m[position-nz*3] += a3x*d[position];  
      m[position-nz*4] += a4x*d[position]; 
      m[position]      += a0z*d[position];
      m[position+1]    += a1z*d[position];
      m[position+2]    += a2z*d[position];
      m[position+3]    += a3z*d[position];  
      m[position+4]    += a4z*d[position];  
      m[position-1]    += a1z*d[position];
      m[position-2]    += a2z*d[position];
      m[position-3]    += a3z*d[position];  
      m[position-4]    += a4z*d[position];   

    }
  }

  return 0;
}

}
#endif
