#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"
#include <omp.h>
#include "base/utilities.h"

#include "base/Vector.h"
#include "base/inCoreFloat.h"
#include "base/DataInCoreFloat.h"
#include "base/Operator.h"
#include "base/MultiOperator.h"

#include "CreateModel.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Prop.h"

using namespace std;
using namespace SEP;

/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/



int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "         Full Waveform inversion.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  long long nx, nz, nt;
  long long sx, sz,ns,ox,dxs;
  long long boundary;
  float dx, dz, dt;
  float velocity, density, modulus;
  float fmax=15;
  nt = 1500;
  nx = 401;
  nz = 401;
  ns = 10;
  sx = nx/2;
  sz = nz/2;
  boundary = 5;

  density=2.;          // [g/cm^3]
  velocity=2.4;        // [km/s]
  dx = 10.; dz = 10.;  // [m]
  dt = 1.5;            // [ms]
  modulus=velocity*velocity*density; // [GPa]

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for shot
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wave field cube
  inCoreFloat3D dspace(ns,nt,nx);         // space for data space
  // Set up vectors
  Vector *K      = new Vector(&mspace);    K->zeroElement();      // migration velocity
  Vector *rho    = new Vector(&mspace);    rho->zeroElement();     // Velocity vector
  Vector *GradK  = new Vector(&mspace);    GradK->zeroElement();      // migration velocity
  Vector *Gradr  = new Vector(&mspace);    Gradr->zeroElement();     // Velocity vector
  Vector *vel    = new Vector(&mspace);    vel->zeroElement();      // migration velocity
  Vector *shot   = new Vector(&sspace);    shot->zeroElement();     // shot profile
  Vector *data   = new Vector(&dspace);    data->zeroElement();     // data
  Vector *u1     = new Vector(&uspace);    u1->zeroElement();       // Wave field
  Vector *u2     = new Vector(&uspace);    u2->zeroElement();       // Wave field  
  Vector *u3     = new Vector(&uspace);    u3->zeroElement();       // Wave field  

  SensorPad *SP = new SensorPad(&dspace, &uspace, sz, boundary);
  SP->DotTest(true); 
  

  // Read in source wavelet and velocity field
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  // Set up model
  Models *CModel =  new Models(nz, nx, modulus, density);
  CModel->ModelC(K,rho);


  // Check numerical condition
  cout << "    Dispersion: Vmin/(fmax*dx) = " << 1000.*velocity/fmax/dx <<", should be > 4"<<endl;
  cout << "    Courant:    Vmax*dt/dx     = " << velocity*dt/dx <<", should be < 0.717"<<endl;
 
  // Generate operators for wave progation
  Prop *Lmig = new Prop(&uspace,&uspace);
  Lmig->SetPropPar(boundary,dt, dx, K);
  Lmig->SetModel(K,rho);
  //Lmig->DotTest(true);
  
  // Inject source wavelet
  Source *Ss = new Source(nt, nz, nx);
  int dims;
  getch("dim","d",&dims);
  if(dims == 2){
    cout << "  2D test " << endl;
    Ss->Injection(shot, u1, sz, sx);
  }
  else if(dims == 1){
    cout << "  1D test " << endl;
    Ss->InjectionLine(shot, u1, sz, sx);
  }
  
  // Forward propagation
  double t0 = timer();
  Lmig->Forward(false,u1,u3);         // source -> ufield
  double t1 = timer(); 
  cout << "   wave propagation, takes " << t1-t0 << "seconds. " <<endl;
  
  float *__restrict__ u3_val = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();
  cout << "wave norm: " << u3->innerProd(*u3) <<endl;

  srite("wave1",u3_val,sizeof(float)*nt*nz*nx);

  // 
  int Eflag;
  getch("Eflag","d",&Eflag);  
  float *err = new float[nt];
  double temp, temp1;
  if(Eflag==1){
    float *pt = new float[nz];
    float *p0 = new float[nz];
    float *pn = new float[nz];
    llong nt0 = 200;
    for(llong iz=0; iz < nz; iz++){
      p0[iz] = u3_val[(nt0-1)*nx*nz+iz*nx+200];
    }
    for(llong it = nt0; it < 700; it+=2){
      // numerical solution
      for(llong iz=0; iz < nz; iz++){
        pn[iz] = u3_val[(it-1)*nx*nz+iz*nx+200];
      }
      // clean pt
      for(llong iz=0;iz<nz;iz++) pt[iz]=0.;
      // theoretical solution
      llong shift = (it-nt0)/2;
      for(llong rz=200+shift; rz < nz; rz++){pt[rz] = p0[rz-shift];}
      for(llong lz=200-shift; lz >= 0; lz--){pt[lz] = p0[lz+shift];}
      // compute norm
      temp = 0;
      temp1=0;
      for(llong iz=0; iz<nz;iz++){
        temp += (pt[iz]-pn[iz])*(pt[iz]-pn[iz]);
        temp1+=pt[iz]*pt[iz];
      }
      cout << " At it=" << it << " , the error = "<< temp << endl;
      err[it] = temp/temp1;
    }  // end of for it loop
    srite("err",err,sizeof(float)*nt);
  }  // end of if


  if(Eflag==4){
    float *ft = new float[nz];
    float *f0 = new float[nz];
    float *fn = new float[nz];
    llong nt0 = 200;
    for(llong iz=0; iz < nz; iz++){
      f0[iz] = u3_val[(nt0-1)*nx*nz+iz*nx+200]*sqrt(abs(iz-200));
    }
    for(llong it = nt0; it < 700; it+=2){
      // numerical solution
      for(llong iz=0; iz < nz; iz++){
        fn[iz] = u3_val[(it-1)*nx*nz+iz*nx+200]*sqrt(abs(iz-200));
      }
	  // clean ft
	  for(llong iz=0;iz<nz;iz++) ft[iz]=0.;
	  // theoretical solution
	  llong shift = (it-nt0)/2;
	  for(llong rz=200+shift; rz < nz; rz++){ft[rz] = f0[rz-shift];}
	  for(llong lz=200-shift; lz >= 0; lz--){ft[lz] = f0[lz+shift];}
      // compute norm
      temp = 0;
      temp1=0;
      for(llong iz=0; iz<nz;iz++){
        temp += (ft[iz]-fn[iz])*(ft[iz]-fn[iz]);
        temp1+=ft[iz]*ft[iz];
      }
      cout << " At it=" << it << " , the error = "<< temp << endl;
      err[it] = temp/temp1;
    }  // end of for it loop
    srite("err",err,sizeof(float)*nt);
  }  // end of if


  // Clean up

   std::cout << "========== Done, Cleaning up ==========" <<std::endl; 

  return 0;
}